/* expr.c
	P.Chongstitvatana
	25 Aug 97
*/

#include "compile.h"

#define  PRIVATE

/* ============================================
   implement the following LL(1) parser
   expr -> term exprs
   exprs -> || term exprs | nil
   term -> f1 terms
   terms -> && f1 terms | nil
   f1 -> f2 f1s
   f1s -> RelOp f2 f1s | nil
   f2 -> f3 f2s
   f2s -> {+,-} f3 f2s | nil
   f3 -> item f3s
   f3s -> {*,/} item f3s | nil
   item -> [- ! * &] identifier
	   [-] number
	   [- !] ( expr )
	   & proc
	   func param

   param is (expr, expr, ... )
   RelOp is < <= == != >= >

   Operators have precedency :
     {*,/} > {+,-} > RelOp > && > ||

   ==================================================
*/


#define  isRelOp(t)   t == tkGT || t == tkGE || t == tkEQEQ || \
		      t == tkNE || t == tkLE || t == tkLT
#define  isUnaryOp(t) t == tkMINUS || t == tkNOT || \
		      t == tkSTAR || t == tkAND

extern sym_entry symtab[];

expr()
{
  term();
  exprs();
}

int chkUnaryOp()
{
  int uop;
  if( isUnaryOp(tok) ) {
    uop = tok;
    lex();
  } else {
    uop = tkNIL;
  }
  return uop;
}

vectorAds( int a )
{
  out2(icLvalg,symtab[a].ref);
  lex();
  expect(tkLBRACKET,"missing [ in index");
  lex();
  expr();
  expect(tkRBRACKET,"missing ] in index");
  out(icIndex);
}

int actualParam()
{
  int pv = 0;
  lex();
  expect(tkLPAREN,"expect (" );
  lex();
  while( tok != tkRPAREN && tok != tkEOF) {
    expr();
    pv++;
    if( tok == tkCOMMA ) lex();
  }
  expect(tkRPAREN,"expect )");
  return pv;
}

PRIVATE exprs()
{
  if( tok == tkOROR ) {
    lex();
    term();
    out(icOr);
    exprs();
  }
}

PRIVATE term()
{
  f1();
  terms();
}

PRIVATE terms()
{
  if( tok == tkANDAND ) {
    lex();
    f1();
    out(icAnd);
    terms();
  }
}

PRIVATE f1()
{
  f2();
  f1s();
}

PRIVATE f1s()
{
  int op;
  if( isRelOp(tok) ) {
    op = tok;
    lex();
    f2();
    switch( op ) {
    case tkLT : out(icLt); break;
    case tkLE : out(icLe); break;
    case tkEQEQ : out(icEq); break;
    case tkNE : out(icNe); break;
    case tkGT : out(icGt); break;
    case tkGE : out(icGe); break;
    }
    f1s();
  }
}

PRIVATE f2()
{
  f3();
  f2s();
}

PRIVATE f2s()
{
  int op;
  if ( tok == tkPLUS || tok == tkMINUS ) {
    op = tok;
    lex();
    f3();
    switch( op ) {
    case tkPLUS : out(icAdd); break;
    case tkMINUS : out(icSub); break;
    }
    f2s();
  }
}

PRIVATE f3()
{
  item();
  f3s();
}

PRIVATE f3s()
{
  int op;
  if( tok == tkSTAR || tok == tkSLASH ) {
    op = tok;
    lex();
    item();
    switch( op ) {
    case tkSTAR : out(icMul); break;
    case tkSLASH : out(icDiv); break;
    }
    f3s();
  }
}

PRIVATE item()
{
  int uop,type,a;
  uop = chkUnaryOp();
  switch( tok ) {
  case tkIDENTIFIER :
    a = chkLocal(tokstring,&type);
    switch( type ) {
    case tySCALAR : doVal(uop,icLvalg,icRvalg,symtab[a].ref); break;
    case tyLOCAL  : doVal(uop,icLval,icRval,symtab[a].ref); break;
    case tyVECTOR : doVector(uop,a); break;
    case tyFUNCTION : doFunc(uop,a); break;
    case tyPROCESS : doProc(uop,a); break;
    }
    lex();
    break;
  case tkNUMBER :
    switch( uop ) {
    case tkNIL : out2(icLit,atoi(tokstring)); break;
    case tkMINUS : out2(icLit,-atoi(tokstring)); break;
    default : seterror("invalid unary op for number");
    }
    lex();
    break;
  case tkLPAREN :
    lex();
    if( tok != tkRPAREN ) {
      expr();
      expect(tkRPAREN,"missing )");
      switch( uop ) {
      case tkMINUS : out2(icLit,0); out(icSub); break;
      case tkNOT : out(icNot); break;
      case tkNIL : break;
      default : seterror("invalid expression");
      }
    }
    lex();
    break;
  }
}

PRIVATE doVal( int uop, int lval, int rval, int ref )
{
  switch( uop ) {
  case tkNIL : out2(rval,ref); break;
  case tkMINUS : out2(rval,ref); out2(icLit,0); out(icSub); break;
  case tkNOT : out2(rval,ref); out(icNot); break;
  case tkAND : out2(lval,ref); break;
  case tkSTAR : out2(rval,ref); out(icFetch); break;
  }
}

PRIVATE doVector( int uop, int a )
{
  vectorAds(a);
  switch( uop ) {
  case tkNIL : out(icFetch); break;
  case tkAND : break;
  case tkSTAR : out(icFetch); out(icFetch); break;
  }
}

PRIVATE doFunc( int uop, int a )
{
  if( uop != tkNIL ) seterror("invalid & at function name");
  if( symtab[a].arg != actualParam() )     /* no. of param */
    seterror("number of actual arg doesn't match declaration");
  out2(icCall,a);
}

PRIVATE doProc( int uop, int a )
{
  if( uop != tkAND ) seterror("missing & in process name");
  out2(icLit,symtab[a].pid);
}

