/* lex.c   lexical analyzer   for R1
	15 Oct 97
	binary search reserved word  		2 Sept 97
   add after, delay, gettime in reserved word	21 Aug 98
*/

#include "lex.h"
#include "compile.h"

#define	PRIVATE	static

int tok;		/* current token type, string */
char tokstring[256];
int line = 1;		/* current position */
uchar CH;       	/* current char */

PRIVATE char chartype[256];	/* character type table */

PRIVATE lookahead(char c, int type, int elsetype);
PRIVATE chkReservedWord( char * str );
PRIVATE int skipblank();

extern  FILE *FI;

initCharType()
{
  int i;
  for(i=0; i<256; i++) chartype[i] = cBLANK;
  for(i='0'; i<='9'; i++ ) chartype[i] = cDIGIT;
  for(i='A'; i<='Z'; i++ ) chartype[i] = cLETTER;
  for(i='a'; i<='z'; i++ ) chartype[i] = cLETTER;
  chartype['_'] = cLETTER;
  chartype['"'] = cQUOTE;
  chartype['*'] = cSTAR;
  chartype['/'] = cSLASH;
  chartype['-'] = cMINUS;
  chartype['+'] = cPLUS;
  chartype['='] = cEQ;
  chartype['&'] = cAND;
  chartype['|'] = cOR;
  chartype['!'] = cNOT;
  chartype['<'] = cLT;
  chartype['>'] = cGT;
  chartype[','] = cCOMMA;
  chartype[';'] = cSEMICOLON;
  chartype['('] = cLPAREN;
  chartype[')'] = cRPAREN;
  chartype['['] = cLBRACKET;
  chartype[']'] = cRBRACKET;
  chartype['{'] = cLBRACE;
  chartype['}'] = cRBRACE;
  chartype[EOF_CHAR] = cEOF;
}

#define isLetterOrDigit(c)  chartype[c] == cLETTER || chartype[c] == cDIGIT
#define isDigit(c)	    chartype[c] == cDIGIT

/* append a char c to string str, length limit at 255 */
#define appendc(str,len,c)  \
  if(len < 256) { str[len] = c; str[len+1] = 0; len++; }
#define setType(t)	    tok = t; getch()

lex()
{
  int t,len;
  t = skipblank();
  if( t == cSLASH ) {
    tok = tkSLASH;
    return;
  }
  len = 0;
  switch( t ) {
  case cLETTER :
	 while( isLetterOrDigit(CH) ) {
      appendc(tokstring,len,CH);
      getch();
    }
    tok = chkReservedWord(tokstring);
    break;
  case cDIGIT :
    while( isDigit(CH) ) {
      appendc(tokstring,len,CH);
      getch();
    }
    tok = tkNUMBER;
    break;
  case cQUOTE :
    getch();
    do {
      appendc(tokstring,len,CH);
      getch();
		if( CH == NL_CHAR ) line++;
    } while ( CH != '"');
    getch();
    tok = tkSTRING;
    break;
  case cSTAR  	: setType(tkSTAR); break;
  case cMINUS  	: setType(tkMINUS ); break;
  case cPLUS   	: setType(tkPLUS ); break;
  case cEQ     	: lookahead('=',tkEQEQ,tkEQ); break;
  case cAND    	: lookahead('&',tkANDAND,tkAND); break;
  case cOR     	: lookahead('|',tkOROR,tkERROR); break;
  case cNOT	: lookahead('=',tkNE,tkNOT); break;
  case cLT	: lookahead('=',tkLE,tkLT); break;
  case cGT     	: lookahead('=',tkGE,tkGT); break;
  case cCOMMA  	: setType(tkCOMMA ); break;
  case cSEMICOLON:setType(tkSEMICOLON ); break;
  case cLPAREN 	: setType(tkLPAREN ); break;
  case cRPAREN 	: setType(tkRPAREN ); break;
  case cLBRACKET: setType(tkLBRACKET ); break;
  case cRBRACKET: setType(tkRBRACKET ); break;
  case cLBRACE 	: setType(tkLBRACE ); break;
  case cRBRACE 	: setType(tkRBRACE ); break;
  case cEOF    	: tok = tkEOF ; break;
  default 	: tok = tkERROR;
  }
}

getch()
{
  int c;
  c = getc(FI);
  if ( c == EOF ) c = EOF_CHAR;
  CH = c;
}

PRIVATE lookahead(char c, int type, int elsetype)
{
  getch();
  if( CH == c ) {
    tok = type;
    getch();
  } else {
    tok = elsetype;
  }
}

#define NKEYWORD 17

/* linear search , ordered by estimate frequency of use --

PRIVATE char keyword[][10] = {
  "if", "else", "while", "print", "return",
  "send", "receive", "wait", "signal", "region",
  "process", "global", "semaphore", "main",0 };

-- if YES instantiate type of token --
PRIVATE isReservedWord2( char * str, int * tktype )
{
  int t;
  for( t=0; t<NKEYWORD; t++ )
    if( strcmp(keyword[t],str) == 0) {
      *tktype = t;
      return 1;
    }
  return 0;
}
*/
/* this code is from Schreiner, A.T. and Friedman, H.G. "Intro. to
   compiler construction with UNIX", Prentice hall, 1985
 */
typedef struct {
  char *name;
  int  tk;
} rw_entry;

PRIVATE rw_entry rwtable[] = {
  "after",      tkAFTER,
  "delay", 	tkDELAY,
  "else", 	tkELSE,
  "gettime",	tkGETTIME,
  "global",     tkGLOBAL,
  "if",         tkIF,
  "main",       tkMAIN,
  "print",      tkPRINT,
  "process",    tkPROCESS,
  "receive",    tkRECEIVE,
  "region",     tkREGION,
  "return",     tkRETURN,
  "semaphore",  tkSEMAPHORE,
  "send",       tkSEND,
  "signal",     tkSIGNAL,
  "wait",       tkWAIT,
  "while",      tkWHILE
};

/* return token of reserved word, if not found return tkIDEN */
PRIVATE int chkReservedWord( char * str )
{
  rw_entry *low = rwtable;
  rw_entry *high = &rwtable[NKEYWORD-1];
  rw_entry *mid;
  int c;

  while( low <= high ) {
    mid = low + (high-low)/2;
    if( ( c = strcmp(mid->name, str)) == 0 )
      return mid->tk;
    else if ( c < 0 )
      low = mid + 1;
    else
      high = mid - 1;
  }
  return tkIDENTIFIER;
}

PRIVATE int skipblank()     /* return chartype of current CH */
{
  for(;;) {
    while( chartype[CH] == cBLANK ) {
      if ( CH == NL_CHAR) line++;
      getch();
    }
    if( CH != '/' ) return(chartype[CH]);
    getch();
    if( CH != '/' ) return(cSLASH);
    while( CH != NL_CHAR && CH != EOF_CHAR) getch();
  }
}


