/* stmt.c
	P. Chongstitvatana
			25 Aug 97
   add after 		22 Aug 98
*/

#include "compile.h"

#define	PRIVATE		static

extern uchar CS[];
extern int CP;
extern int lv;
extern int mainRef;
extern sym_entry symtab[];

pass2()
{
  int a,pv,type,pid,c,adr,i;
  out2(icCall,mainRef);
  lex();
  while( tok != tkEOF ) {
    a = atoi(tokstring);        /* ref to symtab */
    pv = symtab[a].arg;
    type = symtab[a].type;
    pid = symtab[a].pid;
    symtab[a].ref = CP;
    for( i=0; i<pv; i++ ) {     /* formal param */
      lex();
      insertLocal(tokstring);
    }
    if( type == tyFUNCTION ) {
      out3(icFunc,pv,0);
      adr = CP - 2;
    } else {
      out3(icProc,pid,pv);
      adr = CP;
      out1(0);
    }
    lex();
    line = atoi(tokstring);  /* line num */
    lex();
    stmt();
    if( type == tyPROCESS || mainRef == a ) {
      out(icStop);
    } else {
      c = CS[CP - 1];
      if( c != icRet0 && c != icRet1 )
	out(icRet0);
    }
    patch(adr,lv-pv);
    cleanLocal();
  }
}

PRIVATE stmt()
{
  if( tok == tkLBRACE ) {
    lex();
    while( tok != tkRBRACE && tok != tkEOF )
      stmt1();
    expect(tkRBRACE,"missing }");
    lex();
  } else {
    stmt1();
  }
}

PRIVATE stmt1()
{
  switch( tok ) {
  case tkIDENTIFIER : idenStmt(); break;
  case tkSEMICOLON  : lex(); break;
  case tkIF 	    : ifStmt(); break;
  case tkWHILE      : whileStmt(); break;
  case tkRETURN     : returnStmt(); break;
  case tkRBRACE     : break;
  case tkWAIT       : waitStmt(); break;
  case tkSIGNAL     : signalStmt(); break;
  case tkREGION     : regionStmt(); break;
  case tkSEND       : sendStmt(); break;
  case tkRECEIVE    : receiveStmt(); break;
  case tkPRINT      : printStmt(); break;
  case tkDELAY	    : delayStmt(); break;
  case tkGETTIME    : gettimeStmt(); break;
  default           : seterror("invalid statement");
  }
}

PRIVATE idenStmt()
{
  int uop, a, type, ref;
  uop = chkUnaryOp();
  a = chkLocal(tokstring,&type);
  ref = symtab[a].ref;
  switch( type ) {
  case tySCALAR :
  case tySEMAPHORE :
    out2(icLvalg,ref);
    doLHS(uop);
    doRHS();
    break;
  case tyLOCAL  :
    out2(icLval,ref);
    doLHS(uop);
    doRHS();
    break;
  case tyVECTOR :
    vectorAds(a);
	 if( uop == tkSTAR ) out(icFetch);
    doRHS();
    break;
  case tyFUNCTION :
  case tyPROCESS :
    if( actualParam() != symtab[a].arg )
      seterror("number of arg doesn't match declaration");
    lex();
    expect(tkSEMICOLON,"expect ;");
    out2(icCall,a);
    break;
  }
  lex();
}

PRIVATE doLHS( int uop )
{
  switch( uop ) {
  case tkNIL  :  break;
  case tkSTAR : out(icFetch); break;
  default     : seterror("invalid prefix on LHS");
  }
}

PRIVATE doRHS()
{
  lex();
  expect(tkEQ,"missing = in assignment");
  lex();
  expr();
  expect(tkSEMICOLON,"missing ;");
  out(icSet);
}

PRIVATE ifStmt()
{
  int a1, a2;
  lex();
  expr();
  out2(icJz,0);
  a1 = CP - 2;
  stmt();
  if( tok == tkELSE ) {
    out2(icJmp,0);
    a2 = CP - 2;
    patch(a1,CP);
    lex();
    stmt();
    patch(a2,CP);
  } else {
    patch(a1,CP);
  }
}

PRIVATE whileStmt()
{
  int a1, a2;
  lex();
  a1 = CP;
  expr();
  out2(icJz,0);
  a2 = CP - 2;
  stmt();
  out2(icJmp,a1);
  patch(a2,CP);
}

PRIVATE returnStmt()
{
  int a;
  a = CP;
  lex();
  expr();
  if( a == CP )
    out(icRet0);
  else
    out(icRet1);
  expect(tkSEMICOLON,"missing ;");
  lex();
}

PRIVATE signalStmt()
{
  out2(icSignal,doSem());
  expect(tkSEMICOLON,"missing ;");
  lex();
}

PRIVATE regionStmt()
{
  int ref;
  ref = doSem();
  out2(icWait,ref);
  stmt();
  out2(icSignal,ref);
}

PRIVATE sendStmt()
{
  do2Arg();
  if( tok == tkAFTER ) {
    do1Arg();
    out(icTmSend);
    doAfter();
  } else {
    expect(tkSEMICOLON,"missing ;");
    out(icSend);
    lex();
  }
}

PRIVATE receiveStmt()
{
  do2Arg();
  if( tok == tkAFTER ) {
    do1Arg();
    out(icTmReceive);
    doAfter();
  } else {
    expect(tkSEMICOLON,"missing ;");
    out(icReceive);
    lex();
  }
}

PRIVATE waitStmt()
{
  int a;
  a = doSem();
  if( tok == tkAFTER ) {
    do1Arg();
    out2(icTmWait,a);
    doAfter();
  } else {
    expect(tkSEMICOLON,"missing ;");
    out2(icWait,a);
    lex();
  }
}

PRIVATE doAfter()
{
  int a;
  out2(icJmp,0);
  a = CP - 2;
  stmt();
  patch(a,CP);
}

PRIVATE do1Arg()
{
  lex();
  expect(tkLPAREN,"expect (");
  lex();
  expr();
  expect(tkRPAREN,"missing )");
  lex();
}

PRIVATE do2Arg()
{
  int a, h;
  lex();
  expect(tkLPAREN,"expect (");
  lex();
  a = lookup(tokstring, &h);
  if( a == 0 ) seterror("undefined process");
  if( symtab[a].type != tyPROCESS ) seterror("not a process");
  out2(icLit,a);
  lex();
  expect(tkCOMMA,"missing ,");
  lex();
  expr();
  expect(tkRPAREN,"missing )");
  lex();
}

PRIVATE printStmt()
{
  char *s;
  lex();
  expect(tkLPAREN,"expect (");
  lex();
  while ( tok != tkRPAREN && tok != tkEOF ) {
    if ( tok == tkSTRING ) {
      s = tokstring;
      while( *s != '\0' ) {
	if( *s == '\\' && *(s+1) == 'n' ) {
	  out2(icLit,13); out(icPrintch);
	  out2(icLit,10); out(icPrintch);
	  s += 2;
	} else {
	  out2(icLit,*s); out(icPrintch);
	  s++;
	}
      }
      lex();
    } else {
      expr();
      out(icPrint);
    }
    if( tok == tkCOMMA ) lex();
  }
  expect(tkRPAREN,"missing )");
  lex();
  expect(tkSEMICOLON,"missing ;");
  lex();
}

PRIVATE int doSem()
{
  int a, h;
  lex();
  expect(tkLPAREN,"expect (");
  lex();
  a = lookup(tokstring, &h);
  if( a == 0 ) seterror("undefined semaphore");
  if( symtab[a].type != tySEMAPHORE ) seterror("not a semaphore");
  lex();
  expect(tkRPAREN,"missing )");
  lex();
  return symtab[a].ref;
}

PRIVATE delayStmt()
{
  do1Arg();
  expect(tkSEMICOLON,"missing ;");
  out(icDelay);
  lex();
}

PRIVATE gettimeStmt()
{
  do1Arg();
  expect(tkSEMICOLON,"missing ;");
  out(icGettime);
  lex();
}
